/*
   fuse - associative image reconstruction
   Copyright (C) 1997  Scott Draves <spot@cs.cmu.edu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define argi(s,d)   ((ai = getenv(s)) ? atoi(ai) : (d))
#define argd(s,d)   ((ai = getenv(s)) ? atof(ai) : (d))
#define args(s,d)   ((ai = getenv(s)) ? ai : (d))

#include <stdio.h>
#include <math.h>

#include "image.h"


int block, overlap;
int ntries;
double ratio;

/* find difference between two rects.  image1
   is assumed to be the "mask", ie if its pixel
   value is 0, then that pixel doesn't matter
   */
int
compare(image1, x1, y1, image2, x2, y2, nmasked, w, h)
     image *image1, *image2;
     int x1, y1, x2, y2, *nmasked, w, h;
{
  int t, i, j, diff = 0;
  int stride1 = image1->width;
  int stride2 = image2->width;
  int nm = 0;
  uint *pp1 = image1->pixels, *pp2 = image2->pixels;

  CLIP(image1, x1, y1, w, h);
  CLIP(image2, x2, y2, w, h);

  for (i = 0; i < h; i++)
    for (j =0; j < w; j++) {
      uint p1 = pp1[(y1 + i) * stride1 + x1 + j];
      uint p2 = pp2[(y2 + i) * stride2 + x2 + j];
      if (p1) {
	int t;
	t = (int)(p1 & 0xff) - (int)(p2 & 0xff);
	diff += t * t;
	t = (int)((p1 & 0xff00)>>8) - (int)((p2 & 0xff00)>>8);
	diff += t * t;
	t = (int)((p1 & 0xff0000)>>16) - (int)((p2 & 0xff0000)>>16);
	diff += t * t;
      } else {
	nm++;
      }
    }
  *nmasked = nm;
  return diff;
}

void
source_match(image *over,
	     image *images[], int nimages,
	     image *image2,
	     int x, int y)
{
   int nmasked, w = block, h = block;
   int i, best_x, best_y;
   int this_x, this_y;
   double best_diff = -1.0, d;
   image *image1, *best_image;
   CLIP(image2, x, y, w, h);
   if (0 == w || 0 == h)
      return;
   for (i = 0; i < ntries; i++) {
      int denom;
      image1 = images[random() % nimages];
      this_x = random() % (image1->width - w);
      this_y = random() % (image1->height - h);
      d = compare(image2, x, y, image1, this_x, this_y, &nmasked, w, h);
      denom = (w * h - nmasked);
      if (denom == 0) {
	 best_x = this_x;
	 best_y = this_y;
	 best_image = image1;
	 break;
      }
      if (over) {
	 d *= ratio * w * h / (double) denom;
	 d += compare(over, x, y, image1, this_x, this_y, &nmasked, w, h);
      }
      if (d < best_diff || best_diff < 0.0) {
	 best_diff = d;
	 best_x = this_x;
	 best_y = this_y;
	 best_image = image1;
      }
   }
   blit(best_image, best_x, best_y, image2, x, y, w, h);
}


main(int argc, char **argv)
{
   char *ai;
   char *getenv();
   char *out = args("out", "fuse.ppm");
   image *image1, *images[100];
   image *over;
   image *image2;
   int i, nimages;
   int p[1000];
   int bo;

   if (getenv("over")) {
      over = read_image(getenv("over"), 0);
      image2 = new_image(over->width, over->height);
   } else {
      int sz = argi("size", 1000);
      over = 0;
      image2 = new_image(sz, sz);
   }
   ntries = argi("ntries", 10000);
   ratio = argd("ratio", 1.0);
   block = argi("block", 40);
   overlap = argi("overlap", 10);
   bo = block - overlap;
   nimages=0;
   for (i = 1; i < argc; i++) {
      printf("reading %s...\n", argv[i]);
      images[nimages++] = read_image(argv[i], 0);
   }

   srandom(time(0));
   if (getenv("refine")) {
      int i = 0;
      while (1) {
	 int x = random() % (image2->width - block);
	 int y = random() % (image2->height - block);
	 source_match(over, images, nimages, image2, x, y);
      }
   } else if (getenv("scroll")) {
      while (1) {
	 int i, n = image2->width / bo;
	 /* really should do this right */
	 for (i = 0; i < n; i++)
	    p[i] = i;
	 for (i = 0; i < 2*n; i++) {
	    int i1 = random() % n;
	    int i2 = random() % n;
	    int t = p[i1];
	    p[i1] = p[i2];
	    p[i2] = t;
	 }
	 for (i = 0; i < n; i++) {
	    int x = p[i] * bo;
	    int y = image2->height - block - 1;
	    source_match(over, images, nimages, image2, x, y);
	 }
	 /* scroll down a row */
	 memmove(image2->pixels,
		 image2->pixels + image2->width * bo,
		 sizeof(uint) * image2->width *
		 (image2->height - bo));
	 bzero(image2->pixels + image2->width * (image2->height - bo),
	       sizeof(uint) * image2->width * bo);
      }
   } else {
      int r, i, parity = 0;
      int rad;
      if (getenv("round"))
	 rad = ((image2->height < image2->width) ? 
		image2->height : image2->height) / 2;
      else
	 rad = sqrt(image2->width * image2->width +
		      image2->height * image2->height) / 2;
      
      for (r = 0; r < rad; r += bo) {
	 int n = 1 + r * 2.0 * M_PI / bo;
	 if (n > 1000)
	    printf("joiweugunvnsd\n");
	 for (i = 0; i < n; i++)
	    p[i] = i;
	 for (i = 0; i < 2*n; i++) {
	    int i1 = random() % n;
	    int i2 = random() % n;
	    int t = p[i1];
	    p[i1] = p[i2];
	    p[i2] = t;
	 }
	 parity++;
	 for (i = 0; i < n; i++) {
	    int x, y;
	    float theta = 2.0 * M_PI * p[i] / n;
	    if (parity&1)
	       theta += M_PI / n;
	    x = (image2->width - block) / 2 + r * cos(theta);
	    y = (image2->height - block) / 2 + r * sin(theta);
	    source_match(over, images, nimages, image2, x, y);
	    if (0 == i%5) {
	      printf("writing %s...", out);
	      write_image(image2, out);
	      printf("done\n");
	    }
	 }
      }
   }
   printf("writing %s...", out);
   write_image(image2, out);
   printf("done\n");
}
