/*
   fuse - associative image reconstruction
   Copyright (C) 1997  Scott Draves <spot@cs.cmu.edu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define argi(s,d)   ((ai = getenv(s)) ? atoi(ai) : (d))
#define argd(s,d)   ((ai = getenv(s)) ? atof(ai) : (d))
#define args(s,d)   ((ai = getenv(s)) ? ai : (d))

#include <math.h>
#include "image.h"

/* modify the brightness histogram of OVER so that it matches
   that of the images listed on the command line,
   write the remapped image to OUT */

main(int argc, char **argv)
{
   char *ai;
   char *getenv();
   int i, j, x, y;
   int in_total, over_total;
   int in_hist[768];
   int over_hist[768];
   int inverse[1000];
   image *over = read_image(args("over", "over.rgb"), 0);
   image *out = new_image(over->width, over->height);

   for (i = 0; i < 768; i++) {
      in_hist[i] = 0;
      over_hist[i] = 0;
   }
      
   for (i = 1; i < argc; i++) {
      image *in;
      in = read_image(argv[i], 0);
      for (y = 0; y < in->height; y++)
	 for (x = 0; x < in->width; x++) {
	    uint pix = in->pixels[in->width * y + x];
	    int v[3];
	    int index;
	    FRAG(pix, v);
	    /* should use those funny values */
	    index = (v[0] + v[1] + v[2]);
	    in_hist[index]++;
	 }
      free_image(in);
   }

   for (y = 0; y < over->height; y++)
      for (x = 0; x < over->width; x++) {
	 uint pix = over->pixels[over->width * y + x];
	 int v[3];
	 int index;
	 FRAG(pix, v);
	 /* should use those funny values */
	 index = (v[0] + v[1] + v[2]);
	 over_hist[index]++;
      }
   in_total = 0;
   over_total = 0;
   for (i = 0; i < 768; i++) {
      in_total += in_hist[i];
      in_hist[i] = in_total;
      over_total += over_hist[i];
      over_hist[i] = over_total;
   }
   if (1) {
      int prev = 0;
      double r;
      for (i = 0; i < 1000; i++) {
	 r = i / 1000.0;
	 for (; in_hist[prev] / (double) in_total < r; prev++);
	 inverse[i] = prev;
      }
   }

   for (y = 0; y < over->height; y++)
      for (x = 0; x < over->width; x++) {
	 uint pix = over->pixels[over->width * y + x];
	 int v[3];
	 int index, br;
	 double r;
	 FRAG(pix, v);
	 br = (v[0] + v[1] + v[2]);
	 index = 1000 * over_hist[br] / (over_total + 1);
	 if (br) {
	    r = inverse[index] / (double) br;
	    v[0] *= r;
	    v[1] *= r;
	    v[2] *= r;
	 } else {
	    v[0] = v[1] = v[2] = inverse[index];
	 }
	 over->pixels[over->width * y + x] = ASSEMBLE(v[0], v[1], v[2]);
      }
   write_image(over, args("out", "out.rgb"));
}
